/*
	ph_.c ver 1.1 (c) PaulHoule.com  9/13/2019

	Implementation of ph_.lib -- see ph_.h for documentation.

	The approach is to create a C runtime library using thin
	wrappers around Windows OS DLL functions.

	This is optimized for size (not speed or completeness).
	Target use would be a simple Windows command line tool.

	(c) PaulHoule.com
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#pragma comment(lib, "user32")
#pragma comment(lib, "kernel32")

#define PH_LIB_BUILD
#include "ph_.h"

// ph_XF... can't change -- their values are assumed in some formulas.
#define ph_XFBIN 8
#define ph_XFAPP 4
#define ph_XFERR 2
#define ph_XFEOF 1
typedef struct {
	ph_int32_t hw;
	ph_uint16_t err;
	ph_uint8_t f;
	ph_uint8_t u;
} ph_XFILE;
static ph_XFILE ph_xfa[16];			// XFILE array ([1..3] are std err,out,in)

void ph_CC ph_exit(unsigned ecode) {
	ExitProcess(ecode);
}

int ph_CC ph_sprintf(char *pDst, char *pFmt, ...) {
	va_list vl;
	int i;
	va_start(vl, pFmt);
	i = wvsprintf(pDst, pFmt, vl);
	va_end(vl);
	return i;
}

int ph_CC ph_vsprintf(char *pDst, char *pFmt, void *vl) {
	return wvsprintf(pDst, pFmt, (va_list) vl);
}

int ph_CC ph_printf(char *pFmt, ...) {
	char pDst[1040];
	va_list vl;
	int i;
	va_start(vl, pFmt);
	i = wvsprintf(pDst, pFmt, vl);
	va_end(vl);
	return (int) ph_fwrite(pDst, 1, i, ph_stdout);
}

int ph_CC ph_vprintf(char *pFmt, void *vl) {
	char pDst[1040];
	int i;
	i = wvsprintf(pDst, pFmt, (va_list) vl);
	return (int) ph_fwrite(pDst, 1, i, ph_stdout);
}

int ph_CC ph_fprintf(ph_FILE *ph_h, char *pFmt, ...) {
	char pDst[1040];
	va_list vl;
	int i;
	va_start(vl, pFmt);
	i = wvsprintf(pDst, pFmt, vl);
	va_end(vl);
	return (int) ph_fwrite(pDst, 1, i, ph_h);
}

int ph_CC ph_vfprintf(ph_FILE *ph_h, char *pFmt, void *vl) {
	char pDst[1040];
	int i;
	i = wvsprintf(pDst, pFmt, (va_list) vl);
	return (int) ph_fwrite(pDst, 1, i, ph_h);
}

int ph_CC ph_fputc(int c, ph_FILE *ph_h) {
	ph_fwrite((char *) &c, 1, 1, ph_h);
	return c;
}

int ph_CC ph_fputs(char *p, ph_FILE *ph_h) {
	return (int) ph_fwrite(p, 1, ph_strlen(p), ph_h);
}

void *ph_CC ph_malloc(ph_size_t n) {
	return HeapAlloc(GetProcessHeap(), 0, n);
}

void ph_CC ph_free(void *p) {
	if (p) HeapFree(GetProcessHeap(), 0, p);
}

#define ph_mcpy(FN, SRC_DST_T, TMP_T)\
SRC_DST_T *ph_CC FN(SRC_DST_T *pDst, SRC_DST_T *pSrc, ph_size_t n) {\
	ph_size_t nc = n;\
	TMP_T *pD = (TMP_T *) pDst, *pS = (TMP_T *) pSrc;\
	if (nc) do { *pD++ = *pS++; } while (--nc);\
	return pDst;\
}
ph_mcpy(ph_memcpy, void, char)
ph_mcpy(ph_wmemcpy, ph_wchar_t, ph_wchar_t)

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; Had to be careful to write memset in such a way that MSVC compiler
	; didn't replace the loop with a call to memset (which doesn't exist
	; in a build without the MSVC CRT lib).
	*/
#define ph_mset(FN, DST_T, C_T, TMP_T)\
DST_T *ph_CC FN(DST_T *pDst, C_T c, ph_size_t n) {\
	ph_size_t nc = n;\
	if (nc) {\
		TMP_T *pD = pDst;\
		do { *pD++ = (TMP_T) c; } while (--nc);\
	}\
	return pDst;\
}
ph_mset(ph_memset, void, int, char)
ph_mset(ph_wmemset, ph_wchar_t, ph_wchar_t, ph_wchar_t)

#define ph_schr(FN, SRC_T, C_T, TMP_T)\
SRC_T *ph_CC FN(SRC_T *pSrc, C_T c) {\
	TMP_T sc;\
	do {\
		sc = *pSrc++;\
		if (sc == (TMP_T) c) return pSrc - 1;\
	} while (sc);\
	return NULL;\
}
ph_schr(ph_strchr, char, int, char)
ph_schr(ph_wcschr, ph_wchar_t, ph_wchar_t, ph_uint32_t)

#define ph_slen(FN, SRC_T, TMP_T)\
ph_size_t ph_CC FN(SRC_T *pSrc) {\
	SRC_T *pS = pSrc;\
	TMP_T c;\
	do { c = *pS++; } while (c);\
	return (pS - pSrc) - 1;\
}
ph_slen(ph_strlen, char, char)
ph_slen(ph_wcslen, ph_wchar_t, ph_uint32_t)

/*----------------------------------------------------------------------------
;
; Translates external ph_h (which is a ph_xfa index) to a ptr to ph_xfa
; entry.  Also inits entry ->hw for std streams, if not already done.
;
;---------------------------------------------------------------------------*/

static ph_XFILE *ph_CC ph_xlath(ph_FILE *ph_h) {
	ph_XFILE *pX = &ph_xfa[(ph_intptr_t) ph_h];
	if ((ph_size_t) ph_h <= 3			// if need to init in/out/err (1/2/3),
	  && ph_h && !pX->hw) {
			// assumes		: #define STD_INPUT_HANDLE  (DWORD)-10
			// windows		: #define STD_OUTPUT_HANDLE (DWORD)-11
			// definitions	: #define STD_ERROR_HANDLE  (DWORD)-12
		pX->hw = (ph_int32_t) (ph_intptr_t)
		  GetStdHandle((DWORD) -((ph_intptr_t) ph_h + 9));
	}
	return pX;
}

void ph_CC ph_clearerr(ph_FILE *ph_h) {	// clear handle error/eof flags
	ph_xfa[(ph_intptr_t) ph_h].f &= ~(ph_XFERR | ph_XFEOF);
}

int ph_CC ph_ferror(ph_FILE *ph_h) {	// nz if file has error status
	return ph_xfa[(ph_intptr_t) ph_h].f & ph_XFERR;
}

int ph_CC ph_feof(ph_FILE *ph_h) {		// nz if file has EOF status
	return ph_xfa[(ph_intptr_t) ph_h].f & ph_XFEOF;
}

/*----------------------------------------------------------------------------
;
; fopen and _wfopen (wide char _wfopen is Microsoft-specific).
;
;---------------------------------------------------------------------------*/

// sanity check fopen constants; also helps document their use in code
#if GENERIC_READ != 1<<31 || GENERIC_WRITE != 1<<30 \
	|| CREATE_ALWAYS != 2 || OPEN_EXISTING != 3 || OPEN_ALWAYS != 4
#error unexpected windows.h CreateFile-related constant(s)
#endif

#define ph_fopen_fhdr(FN, C_T)	/* generate proto, declare locals */\
ph_FILE * ph_CC FN(C_T *fn, C_T *m) {\
	HANDLE hw;\
	unsigned iX;\
	ph_uint8_t f;

#define ph_fopen_set_f(TMP_T)	/* parses m to set f (flags) */\
	do {\
		TMP_T c = *m;\
		f = (2 << 5) | 1;				/* assume/check-for 'w' */\
		if (c == 'w') break;\
		f = (3 << 5) | 2;				/* assume/check-for 'r' */\
		if (c == 'r') break;\
		f = (4 << 5) | ph_XFAPP | 1;	/* assume/check-for 'a' */\
		if (c == 'a') break;\
		return NULL;\
	} while (0);\
	while (1) {\
		TMP_T c = *++m;\
		if (!c) break;\
		if (c == 'b') { f |= ph_XFBIN; continue; }\
		if (c == '+') { f |= 3; continue; }\
		return NULL;\
	}

#define ph_fopen_set_iX			/* sets iX to 1st free file table index */\
	iX = 3;\
	do {\
		if (++iX == sizeof(ph_xfa) / sizeof(*ph_xfa)) return NULL;\
	} while (ph_xfa[iX].hw);

#define ph_fopen_set_hw(A_or_W)	/* calls CreateFile? to set hw */\
	hw = CreateFile ## A_or_W(\
		fn,									/* file name */\
		(DWORD) f << 30,					/* access mode */\
		FILE_SHARE_READ | FILE_SHARE_WRITE, /* share mode (allow r/w) */\
		NULL,								/* default security attributes */\
		f >> 5,								/* create disposition (2,3,4) */\
		FILE_ATTRIBUTE_NORMAL,				/* flags and attributes */\
		NULL								/* template file */\
	);

#define ph_fopen_ftrl			/* init *pX entry and return */\
	if (hw == INVALID_HANDLE_VALUE) return NULL;\
	ph_xfa[iX].hw = (ph_int32_t) (ph_intptr_t) hw;\
	ph_xfa[iX].f = f & (ph_XFAPP | ph_XFBIN);\
	return (ph_FILE *) (ph_intptr_t) iX;\
}

ph_fopen_fhdr(ph_fopen, char)
	ph_fopen_set_f(char)
	ph_fopen_set_iX
	ph_fopen_set_hw(A)
	ph_fopen_ftrl

ph_fopen_fhdr(ph__wfopen, ph_wchar_t)
	ph_fopen_set_f(ph_uint32_t)
	ph_fopen_set_iX
	ph_fopen_set_hw(W)
	ph_fopen_ftrl


// sanity check mode constants
#if ph__O_TEXT != 0 || ph__O_BINARY != 8 || ph_XFBIN != ph__O_BINARY
#error unexpected _O_BINARY/_O_TEXT/ph_XFBIN constant(s)
#endif

int ph_CC ph__setmode(int fd, int mode) {
	ph_uint8_t *pF = &ph_xfa[fd].f;		// ptr to file table flags
	int rc = *pF;						// rc= current flags (all flags)

	mode = (mode ^ rc) & ph_XFBIN;		// mode= ph_XFBIN set if mode change
	*pF = (ph_uint8_t) (rc ^ mode);		// store new flags (perhaps no change)
	return rc & ph_XFBIN;				// return previous bin/text setting
}

int ph_CC ph_fclose(ph_FILE *ph_h) {
	HANDLE hw;
	ph_XFILE *pX = ph_xlath(ph_h);
	hw = (HANDLE) (ph_size_t) pX->hw;
	*(ph_uint64_t *) pX &= 0;			// zero (re-init) file array entry
	return CloseHandle(hw) ? 0 : ph_EOF;
}

/*----------------------------------------------------------------------------
;
; Filter to handle "\n"->"\r\n" expansion when writing in text mode.
; Always sets *pBio = n, which is invalid if error (zero) return status.
;
;---------------------------------------------------------------------------*/

typedef BOOL (__stdcall *ph_WIN_IO_FN)(
	HANDLE, LPVOID, DWORD, LPDWORD, LPOVERLAPPED);

static BOOL __stdcall ph_WriteText(HANDLE hw, ph_uint8_t *p,
	DWORD n, DWORD *pBio, void *pOv
) {
	ph_uint8_t *pEnd = p + n;			// end of source data (n may be 0)
	DWORD bio;
	BOOL rc;
	ph_uint8_t buf[120];				// block of translated output data

	*pBio = n;
	do {
		ph_uint8_t *pBuf = buf;			// prepare to write translated bytes
		do {
			ph_uint8_t c;
			if (p == pEnd) break;		// break= no more source bytes
			c = *p++;					// translate next source byte
			if (c == '\n') *pBuf++ = '\r';
			*pBuf++ = c;
		} while (pBuf <= &buf[sizeof(buf) - 2]); // loop if more buf room
		rc = WriteFile(hw, buf, (DWORD) (pBuf - buf), &bio, NULL);
	} while (rc && p != pEnd);			// loop in no error and more source
	return rc;
}

/*----------------------------------------------------------------------------
;
; Byte-level i/o (below element-level).  At this point the i/o size has been
; resolved to a DWORD count of bytes, that must NOT be 0 or (DWORD) -1.
;
; Returns count of bytes i/o'd, or (DWORD) -1 for error.
; If returned value < requested bytes, EOF has been hit.
;
; The ph_XFERR or ph_XFEOF bit is set if either even occurs.
;
;---------------------------------------------------------------------------*/

static DWORD ph_CC ph_io_bytes(
	ph_uint8_t *p, DWORD bytes, ph_XFILE *pX, ph_WIN_IO_FN pFn
) {
	DWORD totbio = 0;					// total bytes completed counter
	do {
		DWORD bio;						// bytes i/o'd (return value)
		if (!pFn(						// execute i/o -- if error,
		  (HANDLE) (ph_size_t) pX->hw, p, bytes, &bio, NULL)) {
			DWORD e = GetLastError();	// get Windows error code
			if (e != ERROR_BROKEN_PIPE) { // if not broken pipe (that's EOF),
				pX->err = (ph_uint16_t) e; // save error code
				pX->f |= ph_XFERR;		// set error stream status
				totbio = (DWORD) -1;	// set error return status
				break;
			}
		}
		if (!bio) {						// if 0 bytes returned,
			pX->f |= ph_XFEOF;			// set EOF status
			break;
		}
		totbio += bio;					// update total, ptr, left-to-do
		p += bio;
		bytes -= bio;
	} while (bytes);					// i/o again if incomplete
	return totbio;
}

/*----------------------------------------------------------------------------
;
; Element-level i/o (above byte-level).  At this point we're dealing with an
; i/o of n elements each of size sz bytes.  The i/o could be for 0 bytes.
;
; Note limitation: sz must be < (DWORD) -1 (ie. 0xfffffffe or less).
;
;---------------------------------------------------------------------------*/

static ph_size_t ph_CC ph_io_elems(
	ph_uint8_t *p, ph_size_t sz, ph_size_t n, ph_XFILE *pX, ph_WIN_IO_FN pFn
) {
	ph_size_t teio = 0;					// total elements successfully i/o'd
	ph_size_t mep = 0;					// max elements per i/o

	if (sz) {							// if (perhaps) something to i/o,
		if (sz == (DWORD) sz			// can't handle sz >= (DWORD) -1
		 && (DWORD) sz < (DWORD) -1) {
			mep = (16*1024*1024) / sz;	// calc max elements per i/o
			if (!mep) mep++;			// (must be at least 1)
		}
	}

	do {
		DWORD r;
		if (mep > n) mep = n;			// limit to total elements left
		if (!mep) break;
		r = ph_io_bytes(p, (DWORD) (mep * sz), pX, pFn);
		if (r == (DWORD) -1) break;		// break= i/o error
		p += r;							// advance p in case we're not done
		r /= (DWORD) sz;				// r = count of elements xferred
		teio += r;						// update total elements xferred
		if (r != mep) break;			// break= not full xfer, must be EOF
	} while (n -= mep);

	return teio;
}

ph_size_t ph_CC ph_fread(void *p, ph_size_t sz, ph_size_t n, ph_FILE *ph_h) {
	ph_XFILE *pX = ph_xlath(ph_h);

	if (!(pX->f & ph_XFBIN)) {			// set EOF if (unsupported) text mode
		pX->f |= ph_XFEOF;
	}
	if (pX->f & ph_XFEOF) return 0;		// do nothing EOF status exists
	return ph_io_elems(p, sz, n, pX, ReadFile);
}

ph_size_t ph_CC ph_fwrite(void *p, ph_size_t sz, ph_size_t n, ph_FILE *ph_h) {
	ph_XFILE *pX = ph_xlath(ph_h);

	if (pX->f & ph_XFAPP) {				// if opened as 'a' (append),
		SetFilePointer(					// position to EOF before write
			(HANDLE) (ph_size_t) pX->hw, 0, NULL, FILE_END);
	}
	return ph_io_elems(p, sz, n, pX, pX->f & ph_XFBIN
		? (ph_WIN_IO_FN) WriteFile : (ph_WIN_IO_FN) ph_WriteText);
}

// sanity check offset constants
#if FILE_BEGIN != ph_SEEK_SET || FILE_CURRENT != ph_SEEK_CUR \
	|| FILE_END != ph_SEEK_END
#error unexpected windows.h SetFilePointer-related constant(s)
#endif

int ph_CC ph_fseek(ph_FILE *ph_h, long off, int orig) {
	ph_XFILE *pX = ph_xlath(ph_h);

	pX->f &= ~ph_XFEOF;					// seeking clears EOF state
	SetFilePointer((HANDLE) (ph_size_t) pX->hw, off, NULL, orig);
	return 0;
}

long ph_CC ph_ftell(ph_FILE *ph_h) {
	ph_XFILE *pX = ph_xlath(ph_h);
	return SetFilePointer((HANDLE) (ph_size_t) pX->hw, 0, NULL, ph_SEEK_CUR);
}
