/*
	ph_.h ver 1.1 (c) PaulHoule.com  9/13/2019

	Tiny replacement for stdlib (32 or 64-bit) -- #include this header and
	link with ph_.lib.  Targets are simple x86 PC command line tools.
	Designed for X86 Windows (32 or 64-bit) and Microsoft C compiler.

	Creates smaller .exe.  Also eliminates dependencies -- only the Windows
	DLL's are required (msvcrt.dll is not required).  Speeds up compilation
	since only ph_.h is included, and ph_.h includes no other file.

	A program really needs to be written with the target of using this
	library -- the author needs to appropriately restrict their stdlib use.

	The interface is compatible with (though a very small subset of) stdlib,
	so a program can be built either way for testing, eg:

		#ifdef PH_
		  #include <ph_.h>
		#else
		  #include <stdlib.h>
		  #include <stdio.h>
		  ...etc
		#endif
	
	Some drawbacks (there are many more than these):
	 *)	A very small subset of stdlib is implemented.
	 *) Command line arguments (argc, argv) are not supported -- must
		get command line via Windows GetCommandLine() call.
	 *)	Library is not thread-safe.
	 *) Reading file/stream in text mode is unsupported (yields zero bytes).
		Must fopen(, "rb") or _setmode(, _O_BINARY) to read.
		However text mode write ("\n" --> "\r\n") is supported.
	 *) Fixed file table -- only a dozen fopen'd FILE * handles at a time.
	 *)	What is implemented is only partially functional.  As an example,
		printf() uses the Windows user32.dll wvsprintf() which: (1) doesn't
		support floats; (2) limits the output buffer to 1024 bytes; (3) no
		support for %ll qualifier -- MSFT-specific I64 must be used.
*/

#ifndef PH_H						// include guard
#define PH_H
#ifdef __cplusplus
extern "C" {
#endif

#ifndef PH_LIB_BUILD				// define names available to application
#define size_t		ph_size_t
#define intptr_t	ph_intptr_t
#define ptrdiff_t	ph_ptrdiff_t
#define int8_t		ph_int8_t
#define int16_t		ph_int16_t
#define int32_t		ph_int32_t
#define int64_t		ph_int64_t
#define uint8_t		ph_uint8_t
#define uint16_t	ph_uint16_t
#define uint32_t	ph_uint32_t
#define uint64_t	ph_uint64_t
#define wchar_t		ph_wchar_t
#define wint_t		ph_wint_t
#define NULL		ph_NULL
#define FILE		ph_FILE
#define EOF			ph_EOF
#define _O_TEXT		ph__O_TEXT
#define _O_BINARY	ph__O_BINARY 
#define SEEK_SET	ph_SEEK_SET
#define SEEK_CUR	ph_SEEK_CUR
#define SEEK_END	ph_SEEK_END
#define stdin		ph_stdin
#define stdout		ph_stdout
#define stderr		ph_stderr
#define _fileno(h)	((int) (intptr_t) (h))

#define exit		ph_exit
#define printf		ph_printf
#define vprintf		ph_vprintf
#define sprintf		ph_sprintf
#define vsprintf	ph_vsprintf
#define fprintf		ph_fprintf
#define vfprintf	ph_vfprintf
#define fputs		ph_fputs
#define fputc		ph_fputc
#define malloc		ph_malloc
#define free		ph_free
#define memcpy		ph_memcpy
#define wmemcpy		ph_wmemcpy
#define memset		ph_memset
#define wmemset		ph_wmemset
#define strchr		ph_strchr
#define wcschr		ph_wcschr
#define strlen		ph_strlen
#define wcslen		ph_wcslen
#define _setmode	ph__setmode
#define clearerr	ph_clearerr
#define ferror		ph_ferror
#define feof		ph_feof
#define fopen		ph_fopen
#define _wfopen		ph__wfopen
#define fclose		ph_fclose
#define fread		ph_fread
#define fwrite		ph_fwrite
#define fseek		ph_fseek
#define ftell		ph_ftell
#endif

#define ph_CC __fastcall
#ifdef _WIN64
  typedef unsigned __int64 ph_size_t;
  typedef __int64 ph_intptr_t;
  typedef __int64 ph_ptrdiff_t;
#else
  typedef unsigned __int32 ph_size_t;
  typedef __int32 ph_intptr_t;
  typedef __int32 ph_ptrdiff_t;
#endif
typedef __int8  ph_int8_t;
typedef __int16 ph_int16_t;
typedef __int32 ph_int32_t;
typedef __int64 ph_int64_t;
typedef unsigned __int8  ph_uint8_t;
typedef unsigned __int16 ph_uint16_t;
typedef unsigned __int32 ph_uint32_t;
typedef unsigned __int64 ph_uint64_t;
typedef unsigned __int16 ph_wchar_t;
typedef unsigned __int16 ph_wint_t;

#define ph_NULL ((void *) 0)
typedef struct ph_xfile ph_FILE;
#define ph_EOF (-1)
#define ph__O_TEXT 0
#define ph__O_BINARY 8
#define ph_SEEK_SET 0
#define ph_SEEK_CUR 1
#define ph_SEEK_END 2
#define ph_stdin  ((void *) 1)
#define ph_stdout ((void *) 2)
#define ph_stderr ((void *) 3)

void ph_CC ph_exit(unsigned ecode);

int ph_CC ph_printf(char *pFmt, ...);
int ph_CC ph_vprintf(char *pFmt, void *vl);
int ph_CC ph_sprintf(char *pDst, char *pFmt, ...);
int ph_CC ph_vsprintf(char *pDst, char *pFmt, void *vl);
int ph_CC ph_fprintf(ph_FILE *ph_h, char *pFmt, ...);
int ph_CC ph_vfprintf(ph_FILE *ph_h, char *pFmt, void *vl);
int ph_CC ph_fputs(char *p, ph_FILE *ph_h);
int ph_CC ph_fputc(int c, ph_FILE *ph_h);

void *ph_CC ph_malloc(ph_size_t n);
void ph_CC ph_free(void *p);

void *ph_CC ph_memcpy(void *pDst, void *pSrc, ph_size_t n);
ph_wchar_t *ph_CC ph_wmemcpy(ph_wchar_t *pDst, ph_wchar_t *pSrc, ph_size_t n);
void *ph_CC ph_memset(void *pDst, int c, ph_size_t n);
ph_wchar_t *ph_CC ph_wmemset(ph_wchar_t *pDst, ph_wchar_t c, ph_size_t n);
char *ph_CC ph_strchr(char *pSrc, int c);
ph_wchar_t *ph_CC ph_wcschr(ph_wchar_t *pSrc, ph_wchar_t c);
ph_size_t ph_CC ph_strlen(char *pSrc);
ph_size_t ph_CC ph_wcslen(ph_wchar_t *pSrc);

int ph_CC ph__setmode(int fn, int mode);
void ph_CC ph_clearerr(ph_FILE *ph_h);
int ph_CC ph_ferror(ph_FILE *ph_h);
int ph_CC ph_feof(ph_FILE *ph_h);
ph_FILE * ph_CC ph_fopen(char *fn, char *m);
ph_FILE * ph_CC ph__wfopen(ph_wchar_t *fn, ph_wchar_t *m);
int ph_CC ph_fclose(ph_FILE *ph_h);
ph_size_t ph_CC ph_fread(void *p, ph_size_t sz, ph_size_t n, ph_FILE *ph_h);
ph_size_t ph_CC ph_fwrite(void *p, ph_size_t sz, ph_size_t n, ph_FILE *ph_h);
int ph_CC ph_fseek(ph_FILE *ph_h, long off, int orig);
long ph_CC ph_ftell(ph_FILE *ph_h);
#ifdef __cplusplus
}
#endif
#endif									// include guard
